"""This module implements all mbustypes and their parsing.
The basic mbus data types are mapped onto the python data types where possible.
This means: floats, ints, strings and lists are just the same as in python.
Only mbus-symbols and mbus-data have their own type (class MSymbol) which
inherits UserString and MData which holds its data in self.data."""

from string import split, strip, atof, join, replace
import md5, sha, base64, re, time, UserString
import hmac, mbusif

# util function
isSymbol = re.compile( "^[a-zA-Z][a-zA-Z0-9_\.-]*$" ).match
isFloat  = re.compile( "^[0-9]+\.[0-9]+$" ).match
isInt	 = re.compile( "^[0-9]+$" ).match
isString = re.compile( "^\".*\"$" ).match
isData	 = re.compile( "^\<.*\>$" ).match
isNumber = re.compile( "^[0-9]$" ).match
isAlpha  = re.compile( "^[a-zA-Z]+$" ).match
hashalgo = sha
hashkey  = None

class MMessage:
    def __init__( self, msg_or_dest ):
	self.header = None	# Header information (from, to, ...)
	self.payload = []	# contains list of commands
	if type( msg_or_dest ) == type( "string" ):
	    if msg_or_dest.find( "\n" ) != -1:
		self.parseMessage( msg_or_dest )
	    else:
		self.header = MHeader( maddress( msg_or_dest ) )
	else:
	    self.header = MHeader( msg_or_dest )
    def parseMessage( self, msg ):
	# verify HMAC
	digest = base64.encodestring( hmac.hmac( msg[17:], \
		base64.decodestring( hashkey ), hashalgo ) )[:16]
	if msg[:16] != digest:
	    #print "dig is       :",msg[:16]
	    #print "dig should be:",digest
	    #print "address:",msg[17:100]
	    raise TypeError, "message authentication failed"
	contents = split( msg, "\n" )[:-1]
	header = contents[1]
	payload = contents[2:]
	self.header = MHeader( header )
	for i in payload:
	    self.payload.append( MCommand( i ) )
	#print self.header.srcAdr, self.header.destAdr,
	#for i in self.payload:
	#    print i.name,i.args
    def asString( self ):
	res = self.header.asString() + "\n"
	for i in self.payload:
	    res += i.asString() + "\n"
	digest = base64.encodestring( hmac.hmac( res, \
		base64.decodestring( hashkey ), hashalgo ) )[:16]
	return "%s\n%s"%( digest, res )

class MHeader:
    def __init__( self, header_or_dest ):
	if type( header_or_dest ) == type( "string" ):
	    self.parseHeader( header_or_dest )
	elif type( header_or_dest ) == type( {} ):
	    self.destAdr = header_or_dest
	    self.srcAdr = None # gets filled by MbusIF.send 
	    self.ackList = []
	    mbusif.MbusIF.seqNum += 1
	    self.seqNum = mbusif.MbusIF.seqNum
	    self.reliable = "U"
	    self.version = "mbus/1.0"
	else:
	    raise MbusError( "MHeader constructor failed" )
    def parseHeader( self, header ):
	self.version, self.seqNum, self.timestamp, self.reliable, addrs = \
		split( header, None, 4 )
	self.seqNum = int( self.seqNum )
	self.timestamp = long( self.timestamp )
	srcAdr, destAdr, ackList = split( addrs, ")", 2 )
	srcAdr = strip( srcAdr )[1:]
	destAdr = strip( destAdr )[1:]
	self.srcAdr = maddress( srcAdr )
	self.destAdr = maddress( destAdr )
	ackList = strip( ackList ) 
	#print "DEBUG: src=(%s), dst=(%s), ack=(%s)"%(self.srcAdr,self.destAdr,ackList)
	self.ackList = []
	for a in split( ackList[1:-1] ):
	    self.ackList.append( int( a ) )
    def asString( self ):
	res = ""
	res += self.version + " "
	res += "%d "%( self.seqNum )
	res += "%d "%( time.time() )
	res += self.reliable + " "
	res += "(" + maddress( self.srcAdr ) + ")"
	res += " (" + maddress( self.destAdr ) + ")"
	res += " ("
	for a in self.ackList:
	    res += "%d "%( a )
	res += ")"
	return res

class MCommand:
    def __init__( self, cmdstring_or_cmdname ):
	if cmdstring_or_cmdname.find( " " ) != -1:
	    self.parseCmdString( cmdstring_or_cmdname )
	else:
	    if not isSymbol( cmdstring_or_cmdname ):
		raise TypeError, "MCommand name must be MSymbol ('^[a-zA-Z][a-zA-Z0-9_\.-]*$')"
	    self.name = cmdstring_or_cmdname
	    self.args = []
    def parseCmdString( self, cmdstring ):
	self.name = None	# Name of Command (MSymbol)
	self.name, args = split( cmdstring, None, 1 )
	self.args = parseMList( args )
    def asString( self ):
	return self.name + " " + mlistAsString( self.args )

class MSymbol( UserString.UserString ):
    def __init__( self, string ):
	if not( type(string) == type(self) or isSymbol(string) ):
	    raise TypeError, "Symbol must be '^[a-zA-Z][a-zA-Z0-9_\.-]*$'"
	UserString.UserString.__init__( self, string )
    def asString( self ):
	if not isSymbol( self.data ):
	    raise TypeError, "Symbol must be '^[a-zA-Z][a-zA-Z0-9_\.-]*$'" 
	return self.data

class MData:
    def __init__( self, data = "" ):
	self.data = data
    def asString( self ):
	return "<" + replace( base64.encodestring( self.data )[:-1], "\n", "" ) + ">"

def maddress( string_or_hash ):
    """convert mbus adresses given as string to hash and vice versa"""
    if type( string_or_hash ) == type( "" ):
	res = {}
	for i in split( string_or_hash ):
	    key, value = split( i, ":" )
	    if not isAlpha( key ):
		raise TypeError, "Addresstag must be alpha characters only"
	    for v in value:
		if ord(v) <= 0x21 or ord(v) > 0x7F:
		    raise TypeError, "Addressvalue must be 7bit Ascii, no whitespace, no controlcharacters"
	    res[key] = value
	return res
    elif type( string_or_hash ) == type( {} ):
	res = ""
	for key in string_or_hash.keys():
	    res += "%s:%s "%( key, string_or_hash[key] )
	return res

def parseMList( string ):
    string = strip( string )
    if string[0] != "(" or string[-1] != ")":
	raise MbusError( "parsing exception: expected list, got: " + string )
    string = string[1:-1] + " "
    res     = []	# list of results
    cur     = ""	# current word
    type    = None	# guessed type of current word
    quote   = 0		# 1 if string[-1] == "\" and type == "string"
    brkdpth = 0		# bracket depth
    pos     = 0		# position in string
    size    = len( string )
    while pos < size:
	c = string[pos]
	if type == None:
	    if c == "\"":
		type = "string"
	    elif c == "<":
		type = "data"
	    elif c == "(":
		type = "list"
		brkdpth += 1
	    elif isNumber( c ) or c == "-":
		type = "number"
		cur += c
	    elif isSymbol( c ):
		type = "symbol"
		cur += c
	    elif c == " ":
	    	pass # ignore seperating spaces
	    else:
		raise MbusError( "parsing exception: unknown type" )
	elif type == "symbol":
	    if c == " ":
		res.append( MSymbol( cur ) )
		type = None
		cur = ""
	    else:
		cur += c
	elif type == "string":
	    if c == "\"" and not quote:
		res.append( cur )
		type = None
		cur = ""
	    elif quote:
	        if c == "n":
		    cur += "\n"
		else:
		    cur += c
		quote = 0
	    elif c == "\\":
		quote = 1
	    else:
		cur += c
	elif type == "number":
	    if c == " ":
		res.append( int( cur ) )
		type = None
		cur = ""
	    elif c == ".":
		type = "float"
		cur += c
	    elif isNumber( c ):
		cur += c
	    else:
		raise MbusError( "parsing error: not a number" )
	elif type == "float":
	    if c == " ":
		res.append( atof( cur ) )
		type = None
		cur = ""
	    elif c == "." or isNumber( c ):
		cur += c
	    else:
		raise MbusError( "parsing error: not a float number" )
	elif type == "data":
	    if c == ">":
		res.append( MData( base64.decodestring( cur ) ) )
		type = None
		cur = ""
	    else:
		cur += c
	elif type == "list":
	    if c == ")" and brkdpth == 1:
		res.append( parseMList( "(%s )"%( cur ) ) )
		brkdpth -= 1
	    	type = None
		cur = ""
	    elif c == "(":
		cur += c
		brkdpth += 1
	    elif c == ")":
		cur += c
		brkdpth -= 1
	    else:
		cur += c
	pos += 1
    return res

def quoteString( string ):
    string = replace( string, "\\", "\\\\" )
    string = replace( string, "\"", "\\\"" )
    return replace( string, "\n", "\\n" )

def mlistAsString( mlist ):
    if type( mlist ) == type( "string" ):
	return "\"%s\" "%( quoteString( mlist ) )
    if type( mlist ) == type( 123 ):
	return "%d "%( mlist )
    if type( mlist ) == type( 1.0 ):
	return "%f "%( mlist )
    if type( mlist ) == type( [] ) or type( mlist ) == type( () ):
	res = "("
	for i in mlist:
	    res += mlistAsString( i )
	return res + ")"
    try:
	return mlist.asString() + " "
    except AttributeError:
	raise TypeError, "unsupported datatype in mlist found"

class MbusError:
    """This error class may be raised on internal errors
    (parsing errors, transport errors etc.)"""
    def __init__( self, descr, mmsg = None ):
	self.descr = descr
	self.mmsg = mmsg
